# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import socket
import time
import requests
import requests.exceptions
from typing import Union, List, Optional
from base64 import b64encode
from subprocess import run, CalledProcessError, DEVNULL
from requests.exceptions import ConnectionError
from messidge import default_location
from messidge.client.connection import Connection
from . import TaggedCollection, Taggable, Waitable
from .docker import Docker
from .endpoint import WebEndpoint
from .node import Node
from .send import Sender
from .tunnel import Tunnel
from .volume import Volume
from .container import ExternalContainer


class Location(Waitable):
    """The root location object.

        :param location: An optional fqdn of the location (i.e. tiny.20ft.nz).
        :param location_ip: A optional explicit ip for the broker.
        :param quiet: Set true to not configure logging.
        :param debug_log: Set true to log at DEBUG logging level.
        """

    def __init__(self, location: Optional[str]=None, *, location_ip: Optional[str]=None,
                 quiet: Optional[bool]=False, debug_log: Optional[bool]=False):
        super().__init__()

        # collect parameters
        self.location = location if location is not None else default_location(prefix="~/.20ft")
        self.nodes = {}
        self.volumes = TaggedCollection()
        self.last_best_nodes = None
        self.last_best_node_idx = None
        self.tunnels = {}
        self.endpoints = {}
        self.domains = None
        self.last_heartbeat = time.time()

        # see if we even can connect...
        ip = location_ip if location_ip is not None else self.location
        try:
            run(['ping', '-c', '1', ip], check=True, stdout=DEVNULL)
        except CalledProcessError:
            raise RuntimeError("Cannot ping the requested ip: " + ip)

        # set up logging
        if debug_log and quiet:
            raise ValueError("Can't select both quiet and verbose logging")
        if debug_log or quiet is False:
            logging.basicConfig(level=logging.DEBUG if debug_log else logging.INFO,
                                format='%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s',
                                datefmt='%m%d%H%M%S')

        # connect
        self.conn = Connection(self.location, prefix="~/.20ft", location_ip=location_ip, exit_on_exception=True)
        self.user_pk = self.conn.keys.public_binary()
        self.conn.register_commands(self, Location._commands)
        self.conn.start()
        self.conn.wait_until_ready()  # will throw if the connection had a problem
        self.wait_until_ready()  # doesn't return until a resource offer is made
        self.conn.loop.register_on_idle(self._heartbeat)

    def disconnect(self, container=None, returncode=0):
        """Disconnect from the location - without calling this the object cannot be garbage collected"""
        # the container and returncode are passed if you use 'disconnect' as a termination function on a container
        # i.e. it needs to be there, don't take it off!
        if self.conn is None:  # already disconnected
            return
        logging.info("Disconnecting")
        for endpoint in list(self.endpoints.values()):
            [endpoint.unpublish(cluster) for cluster in list(endpoint.clusters.values())]
        self.endpoints.clear()
        for tunnel in list(self.tunnels.values()):
            tunnel.destroy()
        for node in self.nodes.values():
            for container in list(node.containers.values()):
                for server in list(container.ssh_servers.values()):
                    server.stop()
        self.conn.disconnect()
        self.conn = None

    def node(self) -> Node:
        """Returns a node.

           :return: A node object"""
        return self.ranked_nodes()[0]

    def ranked_nodes(self) -> List[Node]:
        """Ranks the nodes in order of resource availability.

        :return: A list of node objects.

        Note that the difference in processor performance is accounted for and is measured in passmarks."""
        nodes = self.nodes.values()
        if len(nodes) == 0:
            raise ValueError("The location has no nodes")
        return sorted(nodes,
                      key=lambda node: node.stats['cpu'] + node.stats['memory'] - 10 * node.stats['paging'],
                      reverse=True)

    def create_volume(self, tag: Optional[str]=None, async: Optional[bool]=True) -> Volume:
        """Creates a new volume

        :param tag: A globally visible tag (to make the volume globally visible).
        :param async: Enables asynchronous writes.
        :return: The new Volume object.

        Note that asynchronous writes cannot damage a ZFS filesystem although the physical state may lag behind the
        logical state by a number of seconds. Asynchronous ZFS is *very* much faster than synchronous."""
        tag = Taggable.valid_tag(tag)
        msg = self.conn.send_blocking_cmd(b'create_volume', {'user': self.user_pk,
                                                             'tag': tag,
                                                             'async': async})
        logging.info("Created volume: " + msg.uuid.decode())
        vol = Volume(self, msg.uuid, tag)
        self.volumes.add(vol)
        return vol

    def destroy_volume(self, volume: Volume):
        """Destroys an existing volume. This is not a 'move to trash', it will be destroyed.

        :param volume: The volume to be destroyed."""
        if not isinstance(volume, Volume):
            raise TypeError()
        self.conn.send_blocking_cmd(b'destroy_volume', {'user': self.user_pk,
                                                        'volume': volume.uuid})
        logging.info("Destroyed volume: " + volume.uuid.decode())
        self.volumes.remove(volume)

    def all_volumes(self) -> List[Volume]:
        """Returns a list of all volumes on this node.

        :return: A list of Volume objects."""
        return list(self.volumes.values())

    def volume(self, key: Union[bytes, str]) -> Taggable:
        """Return the volume with this uuid, tag or display_name.

        :param key: The uuid or tag of the volume object to be returned.
        :return: A Volume object."""
        return self.volumes.get(self.user_pk, key)

    def endpoint_for(self, fqdn: str) -> WebEndpoint:
        """Return a WebEndpoint for the given fqdn.

        :param fqdn: The fully qualified name the endpoint will represent.
        :return: A WebEndpoint object."""
        for domain, ep in self.endpoints.items():
            if fqdn.endswith(domain):
                return ep
        raise ValueError("There is no endpoint capable of serving: " + fqdn)

    def container_for(self, tag: Union[bytes, str]) -> ExternalContainer:
        """Return a connection onto a container owned by another session, but advertised through a tag.

        :param tag: The tag the container was created with.
        :return: An ExternalContainer object."""
        msg = self.conn.send_blocking_cmd(b'find_tag', {'tag': tag})
        return ExternalContainer(self.conn, msg.params['uuid'], msg.params['node'], msg.params['ip'])

    def ensure_image_uploaded(self, docker_image_id: str, descr: Optional[dict]=None) -> List[str]:
        """Sends missing docker layers to the location.

        :param docker_image_id: use the short form id or name:tag
        :param descr: a previously found docker description
        :return: A list of layer sha256 identifiers

        This is not a necessary step and is implied when spawning a container unless specifically disabled.
        The layers are uploaded on a background thread."""

        # Get a description
        if descr is None:
            descr = Docker.description(docker_image_id, self.conn)
        else:
            # update if an explicit description has been passed
            self.conn.send_cmd(b'cache_description', {'image_id': docker_image_id, 'description': descr})

        # Send the missing layers (if any)
        layers = Sender.layer_stack(descr)
        to_upload = Sender.upload_requirements(layers, self.conn)  # if none, does not need a local docker
        logging.info("Ensuring layers (%d) are uploaded for: %s" % (len(layers), docker_image_id))
        if len(to_upload) > 0:
            logging.info("Layers to upload: %d of %d" % (len(to_upload), len(layers)))
            Sender.send(docker_image_id, to_upload, self.conn)
        return layers

    def _heartbeat(self):
        if time.time() - self.last_heartbeat < 30:
            return
        self.last_heartbeat = time.time()
        self.conn.send_cmd(b'heartbeat')

    def _tunnel_onto(self, container, port, localport, bind, *, timeout=30) -> Tunnel:
        # called from Container
        if isinstance(port, str):
            port = int(port)
        if isinstance(localport, str):
            localport = int(localport)

        # create the tunnel
        container.wait_until_ready()  # otherwise the IP address may not exist on the node and creation will fail
        tunnel = Tunnel(self.conn, container.parent(), container, port, localport, bind, timeout)
        self.tunnels[tunnel.uuid] = tunnel
        tunnel.connect()  # connection done 'late' so we can get the tunnel into tunnels first
        return tunnel

    def _wait_http_200(self, container, dest_port, fqdn, path, localport=None) -> Tunnel:
        # called from Container
        # needs to resolve to localhost because that's where the tunnel will be
        addr = socket.gethostbyname(fqdn)
        if addr != '127.0.0.1':
            raise ValueError("FQDN '%s' does not resolve to localhost" % fqdn)
        logging.info("Waiting on http 200: " + container.uuid.decode())

        # OK
        tnl = self._tunnel_onto(container, dest_port, localport, None)
        logging.debug("Tunnel connected onto: " + container.uuid.decode())

        # poll until it's alive
        url = 'http://%s:%d/%s' % (fqdn, tnl.localport(), path if path is not None else '')
        attempts_remaining = 30
        while True:
            try:
                r = requests.get(url, timeout=2)
                if r.status_code == 200:
                    logging.info("Connected onto: " + url)
                    break
            except (ConnectionError, requests.exceptions.ReadTimeout):
                pass
            attempts_remaining -= 1
            if attempts_remaining == 0:
                raise ValueError("Could not connect to: " + url)
            time.sleep(1)
        return tnl

    def _destroy_tunnel(self, tunnel: Tunnel, container=None, with_command=True):
        # Called from Container
        if container is not None:
            if tunnel.container() != container:
                raise ValueError("Tried to destroy a tunnel actually connected to a different container")
        tunnel.destroy(with_command)
        del self.tunnels[tunnel.uuid]

    def _from_proxy(self, msg):
        try:
            tunnel = self.tunnels[msg.uuid]
        except KeyError:
            logging.debug("Data apparently from an already removed tunnel (dropped)")
            return
        try:
            tunnel.from_proxy(msg)
        except KeyError:
            logging.debug("Data arrived from a proxy we seemingly already closed")

    def _close_proxy(self, msg):
        try:
            tunnel = self.tunnels[msg.uuid]
        except KeyError:
            logging.debug("Asked to close a proxy on an already removed tunnel (dropped)")
            return
        try:
            tunnel.close_proxy(msg)
        except KeyError:
            logging.debug("Asked to close a proxy that we already closed")

    def _resource_offer(self, msg):
        self.endpoints = {dom['domain']: WebEndpoint(self.conn, dom['domain']) for dom in msg.params['domains']}
        self.nodes = {node[0]: Node(self, node[0], self.conn, node[1]) for node in msg.params['nodes']}
        self.volumes = TaggedCollection([Volume(self, vol['uuid'], vol['tag']) for vol in msg.params['volumes']])

        self.mark_as_ready()  # only ready once we've dealt with the resource offer

    def _update_stats(self, msg):
        if msg.params['node'] not in self.nodes:
            logging.debug("Received updated stats from a node we didn't know existed: " +
                          b64encode(msg.params['node']).decode())
            return
        node = self.nodes[msg.params['node']]
        node._update_stats(msg.params['stats'])
        self.last_best_nodes = None  # force a refresh next time 'best node' is called

    def _log(self, msg):
        if msg.params['error']:
            logging.error(msg.params['log'])
        else:
            logging.info(msg.params['log'])

    _commands = {b'resource_offer': ([], False),
                 b'update_stats': (['node', 'stats'], False),
                 b'from_proxy': (['proxy'], False),
                 b'close_proxy': (['proxy'], False),
                 b'log': (['error', 'log'], False)}

    def __repr__(self):
        return "<tfnz.location.Location object at %x (nodes=%d)>" % (id(self), len(self.nodes))



# Copyright (c) 2017 David Preece, All rights reserved.
# 
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
import requests.exceptions
import json
import logging
from typing import Optional


class Docker:
    docker_socket = '/var/run/docker.sock'
    docker_url_base = 'http+unix://%2Fvar%2Frun%2Fdocker.sock'
    session = None

    @staticmethod
    def description(docker_image_id: str, conn: Optional['Connection']=None) -> dict:
        """Describe a docker image.

        :param docker_image_id: Docker image id.
        :param conn: An optional connection to the location.
        :return: A dict representation of image metadata."""
        # try locally
        can_connect_local = True
        try:
            r = Docker._session().get('%s/images/%s/json' % (Docker.docker_url_base, docker_image_id))

            # local docker works but doesn't have it
            if r.status_code == 404:
                logging.info("Local docker doesn't have image, trying for remote")
            else:
                # presumably worked, cache it and return
                descr = json.loads(r.text)

                # strip some stuff we don't need
                removes = ('Container', 'Comment', 'ContainerConfig', 'GraphDriver')
                for remove in removes:
                    if remove in descr:
                        del descr[remove]

                # all good
                return descr
        except requests.exceptions.ConnectionError:
            can_connect_local = False

        # no go locally, try remotely
        if conn is not None:
            logging.info("Retrieving description: " + docker_image_id)
            msg = conn.send_blocking_cmd(b'retrieve_description', {'image_id': docker_image_id})
            if 'description' in msg.params:
                return msg.params['description']

        # image is in neither location
        if can_connect_local:
            raise RuntimeError("Cannot find image in either local docker or remote image cache: " + docker_image_id)
        else:
            # local's dead, too
            Docker._docker_warning()

    @staticmethod
    def tarball(docker_image_id: str) -> bytes:
        """Retrieve the tarball of a docker image.

        :param docker_image_id: Docker image id.
        :return: A stream of bytes that would be the contents of the tar archive."""
        try:
            r = Docker._session().get('%s/images/%s/get' % (Docker.docker_url_base, docker_image_id))
            return r.content
        except requests.exceptions.ConnectionError:
            Docker._docker_warning()


    @staticmethod
    def last_image() -> str:
        """Finding the most recent docker image on this machine.

        :return: Docker image id of the most recently built docker image"""
        r = None
        try:
            r = Docker._session().get('%s/images/json' % Docker.docker_url_base)
        except requests.exceptions.ConnectionError:
            Docker._docker_warning()
        if len(r.text) == 0:
            raise ValueError("Docker has no local images.")
        obj = json.loads(r.text)
        return obj[0]['Id'][7:19]

    @staticmethod
    def _docker_warning():
        print("""
    Cannot (and need to) connect to the docker socket.
    
    --------------------------------------
    Is docker running? You may need to run
    sudo chmod 666 /var/run/docker.sock
    --------------------------------------
    """, file=sys.stderr)
        raise RuntimeError("Need a functioning local Docker")

    @staticmethod
    def _session():
        # when we need unix sockets (not deployed on server, hence late binding)
        if Docker.session is None:
            import requests_unixsocket
            Docker.session = requests_unixsocket.Session()
        return Docker.session



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import weakref
import logging
from typing import List, Optional, Tuple
from tfnz.container import Container


class Cluster:
    """An object representing a collection of containers, load balanced and published to an endpoint"""

    def __init__(self, containers: List[Container], rewrite: Optional[str]=None):
        self.uuid = None
        self.containers = {}
        self.rewrite = rewrite
        for container in containers:
            container.wait_until_ready()
            self._add(container)

    def uuids(self):
        return self.containers.keys()

    def _add(self, container):
        if container.uuid in self.containers:
            pass
        self.containers[container.uuid] = container

    def __repr__(self):
        return "<tfnz.endpoint.Cluster object at %x (containers=%d)>" % (id(self), len(self.containers))


class WebEndpoint:
    """An HTTP proxy that can expose a number of clusters"""

    def __init__(self, conn: 'Connection', domain: str):
        """Do not construct directly, see location.endpoints"""
        self.conn = weakref.ref(conn)
        self.domain = domain
        self.domainchars = len(domain)
        self.clusters = {}  # uuid to cluster

    def publish(self, cluster: Cluster, fqdn: str, *, ssl: Optional[Tuple]=None):
        # checks
        if not fqdn.endswith(self.domain):
            raise ValueError("Web endpoint for (%s) cannot publish: %s" % (self.domain, fqdn))
        subdomain = fqdn[:-self.domainchars]

        # ssl check and read
        combined = None
        if ssl is not None:
            if len(ssl) not in (2, 3):
                raise ValueError("SSL needs to be a tuple of (cert.pem, key.pem) or "
                                 "(cert.pem, key.pem, cert.intermediate")
            combined = ''
            with open(ssl[0]) as f:
                combined += f.read()
            with open(ssl[1]) as f:
                combined += f.read()
            if len(ssl) is 3:
                with open(ssl[2]) as f:
                    combined += f.read()

        # tell the location to publish
        msg = self.conn().send_blocking_cmd(b'publish_web', {'domain': self.domain,
                                                             'subdomain': subdomain,
                                                             'rewrite': cluster.rewrite,
                                                             'ssl': combined,
                                                             'containers': list(cluster.uuids())})
        logging.info("Published (%s) at: %s" % (msg.uuid.decode(), subdomain + self.domain))
        cluster.uuid = msg.uuid
        self.clusters[msg.uuid] = cluster
        return msg.uuid

    def unpublish(self, cluster: Cluster):
        self.conn().send_cmd(b'unpublish_web', {'uuid': cluster.uuid})
        logging.info("Unpublished: " + cluster.uuid.decode())
        del self.clusters[cluster.uuid]

    def __repr__(self):
        return "<tfnz.endpoint.WebEndpoint object at %x (domain=%s publishing=%s)>" % \
               (id(self), len(self.domain), ' '.join([c for c in self.clusters.keys()]))



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# the method to implement is tf_main(location, node, environment, portmap, pass_args)

import argparse
import sys
import signal
import re
import os
import os.path
from sys import exit, argv, stderr
from messidge import default_location, create_account
from tfnz.location import Location
from tfnz.volume import Volume
from tfnz.docker import Docker
from tfnz.endpoint import Cluster
from tfnz.cli import base_argparse, systemd, Interactive


def main_impl():
    # see if we have an account
    try:
        default_location(prefix="~/.20ft")
    except RuntimeError:
        # there is no default location
        if len(argv) != 3:
            print("""
There is no a 20ft account on this machine, you need to request 
access from your administrator.

If you already have an account on another machine you can merely 
copy the directory ~/.20ft (and it's contents) to this machine.""", file=stderr)
            return None

        # trying to create a new account using token passed by the server
        create_account(argv[1], argv[2], prefix="~/.20ft")
        print("Created OK for: %s\n" % argv[1])
        return None

    # right, get on with it
    parser = base_argparse('tf')
    launch_group = parser.add_argument_group('launch options')
    launch_group.add_argument('-v', help='verbose logging', action='store_true')
    launch_group.add_argument('-q', help='no logging', action='store_true')
    launch_group.add_argument('-i', help='interactive', action='store_true')
    launch_group.add_argument('-e', help='set an environment variable', action='append', metavar='VAR=value')
    launch_group.add_argument('-f', help='write a pre-boot file', action='append', metavar='src:dest')
    launch_group.add_argument('-m', help='mount a volume', action='append', metavar='uuid:mountpoint')
    launch_group.add_argument('-p', help='add a local->remote tcp proxy', action='append', metavar='8080:80')
    launch_group.add_argument('-c', help='use this command (in container) to start', metavar='script.sh')
    launch_group.add_argument('-w', help='publish on web endpoint',
                              metavar='subdomain.my.com[:www.my.com[:certname]]')
    interactive_group = parser.add_argument_group('local/interactive options')
    interactive_group.add_argument('--ssh', help='create an ssh/sftp wrapped shell on given port', metavar='2222')
    interactive_group.add_argument('-s', help='short form for "--ssh 2222"', action='store_true')
    interactive_group.add_argument('-z', help='launch the container asleep (instead of cmd)', action='store_true')
    server_group = parser.add_argument_group('server options')
    server_group.add_argument('--systemd', help='create a systemd service', metavar='user@tinyserver.my.com')
    server_group.add_argument('--identity', help='specify an identity file')

    parser.add_argument('source', help="if '.', runs the most recently added docker image; "
                                       "else this is the tag or hex id of an image to run.")
    parser.add_argument('args', help='arguments to pass to a script or subprocess (you may need "--", see man page)',
                        nargs=argparse.REMAINDER)
    args = parser.parse_args()

    # collect any pre-boot files
    preboot = []
    if args.f is not None:
        for e in args.f:
            match = re.match('^[\w:]*[\w.\-\\/]+:[\w.\-\\/]+$', e)
            if not match:
                print("Pre-boot copies need to be in source:destination pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return None
            files = e.split(':')
            # are we copying into a directory? append the original filename
            if files[1][-1:] == '/':
                files[1] += files[0]
            # try to read the file in
            try:
                with open(files[0], 'rb') as f:
                    preboot.append((files[1], f.read()))
            except FileNotFoundError:
                print("Could not find the source pre-boot file: " + files[0], file=sys.stderr)
                return None

    # if publishing to an endpoint, we might need ssl certs
    endpoint = None
    rewrite = None
    cert = None
    fqdns = None
    if args.w is not None:
        # split into endpoint:rewrite:certname (rewrite and certname are optional)
        fqdns = args.w.split(':')
        if len(fqdns) == 0 or len(fqdns[0]) == 0:
            print("Cannot publish to an endpoint without an address", file=sys.stderr)
            return None
        endpoint = fqdns[0]

        if len(fqdns) > 1 and len(fqdns[1]) > 0:
            rewrite = fqdns[1]

        if len(fqdns) > 2 and len(fqdns[2]) > 0:
            cert = (fqdns[2] + '.crt', fqdns[2] + '.key')
            if not os.path.exists(cert[0]):
                print("Cannot find certificate for ssl: " + cert[0], file=sys.stderr)
                return None
            if not os.path.exists(cert[1]):
                print("Cannot find key for ssl: " + cert[1], file=sys.stderr)
                return None

    # connect
    location = None
    try:
        location = Location(args.location, location_ip=args.local, quiet=args.q, debug_log=args.v)
    except BaseException as e:
        print("Failed while connecting to location: " + str(e), file=sys.stderr)
        return location

    # are we making a systemd service?
    if args.systemd is not None:
        systemd(location, args, argv, preboot, cert)
        return location

    # are we going to be using the most recent build?
    if args.source == '.':
        args.source = Docker.last_image()

    # maybe remove --
    if len(args.args) > 0:
        if args.args[0] == '--':
            del args.args[0]

    # create env-vars
    e_vars = set()
    environment = []
    if args.e is not None:
        for e in args.e:
            match = re.match('[0-9A-Za-z:_]+=', e)
            if not match:
                print("Environment variables need to be passed as 'name=value' pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            variable = match.group(0)[:-1]
            if variable in e_vars:
                print("Can only pass one value per environment variable.", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            value = e[len(variable)+1:]
            e_vars.add(variable)
            environment.append((variable, value))

    # create portmaps
    l_ports = set()
    portmap = []
    if args.p is not None:
        for e in args.p:
            match = re.match('\d+:\d+$', e)
            if not match:
                print("Portmaps need to be passed as number:number pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            local, remote = e.split(':')
            if local in l_ports:
                print("Cannot bind a local port twice.", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            l_ports.add(local)
            portmap.append((local, remote))

    # create volume mounts ensuring they don't overlap
    volumes = []
    mount_points = set()
    if args.m is not None:
        for m in args.m:
            if ':' not in m:
                print("Volumes need to be passed as uuid:mountpoint pairs", file=sys.stderr)
                print("....error in '%s'" % m, file=sys.stderr)
                return location
            find = m.rfind(':')
            key = m[:find]
            mount = m[find+1:]
            intersection = Volume.trees_intersect(mount_points, mount)
            if intersection is not None:
                print("Error in volumes: %s is a subtree of %s" % (intersection[0], intersection[1]), file=sys.stderr)
                return location
            try:
                volumes.append((location.volumes.get(location.user_pk, key), mount))
            except KeyError:
                print("Could not find volume: " + key)
                return location
            mount_points.add(mount)

    # try to launch the container
    interactive = Interactive(location) if args.i else None
    try:
        node = location.node()
        container = node.spawn_container(args.source,
                                         env=environment,
                                         pre_boot_files=preboot,
                                         volumes=volumes,
                                         stdout_callback=(interactive.stdout_callback if args.i
                                                          else lambda _, out: sys.stdout.buffer.write(out)),
                                         termination_callback=(interactive.termination_callback if args.i
                                                               else location.disconnect),
                                         command=args.c,
                                         sleep=args.z)
        container.wait_until_ready()  # a transport for exceptions
    except BaseException as e:
        print("Failed while spawning container: " + str(e), file=sys.stderr)
        return location

    # create the tunnels
    for m in portmap:
        container.attach_tunnel(m[1], m[0])

    # launch an ssh server? user/pass are anything/anything - you can only connect from localhost anyway
    if args.ssh or args.s:
        container.create_ssh_server(2222 if args.s else int(args.ssh))

    # publish to an endpoint?
    if args.w is not None:
        container.wait_until_ready()
        clstr = Cluster([container], rewrite=rewrite)
        ep = location.endpoint_for(endpoint)
        ep.publish(clstr, fqdns[0], ssl=cert)

    # wait until quit
    if args.i:
        interactive.stdin_loop(container)
    else:
        try:
            location.conn.wait_until_complete()
        except (KeyboardInterrupt, AttributeError):
            pass
    return location


def main():
    maybe_location = main_impl()
    if maybe_location is not None:
        maybe_location.disconnect()


if __name__ == "__main__":
    main()



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import weakref
import socket
import os
import os.path
import paramiko
import paramiko.rsakey
import paramiko.ssh_exception
import logging
import shortuuid
import re
from threading import Thread
from .sftp import Sftp
from . import Waitable
from .tunnel import Tunnel


class SshServer(Waitable):
    """An SSH server using 20ft SDK calls.
    Username/password are whatever you want.
    Note that for the purposes of this, tunnels are regarded as processes because it makes loads of things simpler.

    Do not instantiate directly, use container.create_ssh_server"""

    def __init__(self, container, port):
        Waitable.__init__(self)
        self.uuid = shortuuid.uuid()
        self.container = weakref.ref(container)
        self.node = weakref.ref(container.parent())
        self.location = weakref.ref(self.node().parent())
        self.port = port
        self.run_thread = None
        self.stopping = False
        self.transports = {}  # id to object

        # get a host key
        host_key_fname = os.path.expanduser('~/.20ft/host_key')
        try:
            self.host_key = paramiko.RSAKey.from_private_key_file(host_key_fname)
        except FileNotFoundError:
            self.host_key = paramiko.rsakey.RSAKey.generate(1024)
            self.host_key.write_private_key_file(host_key_fname)

        # see if we can bring the socket up
        self.sock = socket.socket()
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.sock.bind(('', self.port))
        except OSError:
            self.unblock_and_raise(RuntimeError("Tried to start ssh server but the port was already taken: "
                                                + str(self.port)))

        # we are good to go
        self.sock.listen()
        logging.info("SSH server listening: ssh -p %s root@localhost" % self.port)

    def start(self):
        # creates the run loop on a separate thread
        self.run_thread = Thread(target=self.run, name="SSH server onto: " + self.container().uuid.decode())
        self.run_thread.start()
        self.mark_as_ready()

    def stop(self):
        # close the transports
        for transport in list(self.transports.values()):
            transport.stop()

        # close our own accept loop
        self.stopping = True
        self.run_thread.join()
        self.sock.close()

    def run(self):
        # an accept loop
        self.sock.settimeout(0.5)  # so we can poll for self.stopping
        while not self.stopping:
            try:
                # accept a new connection or poll
                try:
                    client, addr = self.sock.accept()
                except socket.timeout:
                    continue

                # wrap a transport round it
                transport = SshTransport(self, client, self.host_key)
                self.transports[transport.uuid] = transport

            except EOFError:
                raise ValueError("There was a problem with ssh - Is there an old key in 'known_hosts'?")

        logging.debug("Exiting server accept loop")

    def transport_closed(self, transport):
        logging.debug("SSH server knows transport has closed: " + transport.uuid)
        del self.transports[transport.uuid]


class SshTransport(paramiko.ServerInterface):
    """A Transport is the per-client abstraction, it will spawn channels."""

    def __init__(self, parent, socket, host_key):
        self.uuid = shortuuid.uuid()
        self.parent = weakref.ref(parent)
        self.container = weakref.ref(parent.container())
        self.lead_channel = None
        self.socket = socket
        self.paramiko_transport = paramiko.Transport(socket)
        self.paramiko_transport.add_server_key(host_key)
        self.paramiko_transport.set_subsystem_handler('sftp', paramiko.SFTPServer, Sftp)
        self.paramiko_transport.start_server(server=self)  # for this one connection

        # channels and port forwarding
        self.channels = {}  # ssh channel id to SshChannel
        self.reverses = {}  # port to SshReverse

        # tunnels are requested before the channel is open
        self.pending_tunnels = {}  # ssh channel id to list of port numbers
        self.pending_reverses = []  # a list of reverse objects

        # start the transport's accept loop
        self.accept_thread = Thread(target=self.channel_accept, name="SSH transport: " + self.uuid)
        self.accept_thread.start()

    def channel_accept(self):
        # a per-transport accept loop
        close_callback = self.stop
        while True:
            channel = self.paramiko_transport.accept()

            # is the transport closing down?
            if channel is None:
                break  # breaks the loop

            # otherwise we're all good
            chid = channel.get_id()
            logging.debug("Accepted paramiko channel: " + str(chid))
            ssh_channel = SshChannel(channel, self.container(), close_callback)
            if close_callback is not None:  # is the leader channel
                self.lead_channel = ssh_channel
            self.channels[chid] = ssh_channel
            close_callback = None

            # waiting to spawn forward tunnels?
            if chid in self.pending_tunnels:
                for port in self.pending_tunnels[chid]:
                    logging.debug("Spawning pending forward for port: " + str(port))
                    ssh_channel.spawn_tunnel(port)  # can only happen after .accept has happened
                del self.pending_tunnels[chid]

            # waiting to spawn channels for reverse tunnels?
            # keeps track of it's own child tunnels
            for reverse in self.pending_reverses:
                try:
                    logging.debug("Hooking up pending reverse for port: " + str(reverse.dest_addr[1]))
                    reverse.spawn_channel()
                    self.reverses[reverse.dest_addr[1]] = reverse
                except ValueError as e:
                    logging.warning(e)
            self.pending_reverses.clear()

        # exited the loop
        logging.debug("Transport loop exited for: " + self.uuid)

        # let the parent know we're done
        self.parent().transport_closed(self)

    def stop(self):
        if self.socket is None:
            return

        # do reverses first because they will hold open processes in sessions
        for reverse in list(self.reverses.values()):
            reverse.close()
        for channel in list(self.channels.values()):
            channel.close_callback = None
            channel.close()

        # close down the transport
        self.socket.close()
        self.accept_thread.join()
        self.paramiko_transport.close()
        self.socket = None
        logging.debug("Transport closed for: " + self.uuid)

    # Paramiko authentication (or lack thereof)
    def check_auth_none(self, username):
        return paramiko.AUTH_SUCCESSFUL

    def get_allowed_auths(self, username):
        return 'none'

    def check_channel_request(self, kind, chanid):
        logging.debug("SSH channel_request (%s): %d" % (kind, chanid))
        if kind == 'session':
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    # Tunnel requests (-L) are requested before the channel is opened
    # We keep track of what we agreed to and open the forwarding requests once the channel has been accepted
    def check_channel_direct_tcpip_request(self, chanid, origin, destination):
        if chanid in self.channels:
            logging.info("SSH direct_tcpip_request received on open channel, just returning 'succeeded'")
        else:
            logging.info("SSH direct_tcpip_request waiting on port: " + str(destination[1]))
            if chanid not in self.pending_tunnels:
                self.pending_tunnels[chanid] = []
            self.pending_tunnels[chanid].append(destination[1])
        return paramiko.OPEN_SUCCEEDED

    # These are the reverse forwarding requests (-R)
    # You need to open a channel back to the client but the channel from the client needs to have been opened first
    # hence we just keep track of this, say "yes" and then wait for the client channel to be accepted.
    # it doesn't work to check for the reverse being open at this point because it could be, under another transport
    def check_port_forward_request(self, address, port):
        logging.info("SSH port_forward_request (reverse): %s:%d" % (address, port))
        self.pending_reverses.append(SshReverse((address, port), self.container(), self))
        return port

    # Callbacks across the server interface
    def check_channel_pty_request(self, channel, term, width, height, pixelwidth, pixelheight, modes):
        # an occasional race condition
        while not channel.get_id() in self.channels:
            os.sched_yield()
        self.channels[channel.get_id()].set_pty_dimensions(width, height)
        return True

    def check_channel_window_change_request(self, channel, width, height, pixelwidth, pixelheight):
        # gets called if window changes size while server is active
        self.channels[channel.get_id()].set_pty_dimensions(width, height)
        self.channels[channel.get_id()].send_window_change()
        return True

    def check_channel_exec_request(self, channel, command):
        while not channel.get_id() in self.channels:
            os.sched_yield()
        return self.channels[channel.get_id()].spawn_process(command)

    def check_channel_shell_request(self, channel):
        while not channel.get_id() in self.channels:
            os.sched_yield()
        return self.channels[channel.get_id()].spawn_shell()


class SshChannel:
    """A channel is a single channel through a transport. It can be connected to only one process."""

    def __init__(self, paramiko_channel, container, close_callback=None):
        self.paramiko_channel = paramiko_channel
        self.container = weakref.ref(container)
        self.connection = weakref.ref(container.conn())
        self.node = weakref.ref(container.parent())
        self.location = weakref.ref(self.node().parent())
        self.loop = weakref.ref(self.connection().loop)
        self.width = 80
        self.height = 24
        self.process = None  # only one process
        self.close_callback = close_callback

    def close(self, obj=None, returncode=0):  # signature needs to be this for callback to work
        # if we've been closed already, just skip out
        if self.paramiko_channel is None:
            return

        # has this been created with a callback?
        if self.close_callback is not None:
            logging.debug("Closing lead channel.")
            cc = self.close_callback
            self.close_callback = None
            cc()
            return

        # close the channel
        chid = self.paramiko_channel.get_id()
        fd = self.paramiko_channel.fileno()
        logging.debug("[chan %d] closing" % chid)

        self.paramiko_channel.send_exit_status(returncode)
        self.paramiko_channel.shutdown(2)
        self.paramiko_channel.close()
        self.paramiko_channel = None
        if fd in self.loop().exclusive_handlers:
            self.loop().unregister_exclusive(fd)

        # close the process (they aren't tied into the container in this case)
        if self.process is not None:
            if self.process.uuid in self.location().tunnels:
                del self.location().tunnels[self.process.uuid]
            self.process.destroy()

    def get_id(self):
        return self.paramiko_channel.get_id()

    # events coming in from the transport

    def set_pty_dimensions(self, width, height):
        self.width = width
        self.height = height

    def send_window_change(self):
        # send the window change event
        self.connection().send_cmd(b'tty_window', {'node': self.node().pk,
                                                   'container': self.container().uuid,
                                                   'process': self.process.uuid,
                                                   'width': self.width,
                                                   'height': self.height})

    def spawn_process(self, command):
        # spawn the process on the end of this channel
        self.process = self.container().spawn_process(command.decode(),
                                                      data_callback=self.data,
                                                      stderr_callback=self.stderr,
                                                      termination_callback=self.close)
        self.loop().register_exclusive(self.paramiko_channel.fileno(), self.event,
                                       comment="SSH Process " + self.process.uuid.decode())
        return True

    def spawn_shell(self):
        # spawn the process on the end of this channel
        self.process = self.container().spawn_shell(data_callback=self.data,
                                                    termination_callback=self.close,
                                                    echo=True)
        self.loop().register_exclusive(self.paramiko_channel.fileno(), self.event,
                                       comment="SSH Shell " + self.process.uuid.decode())
        self.send_window_change()
        return True

    def spawn_tunnel(self, port):
        logging.debug("Spawning forwarding tunnel for port: " + str(port))
        tunnel = Tunnel(self.connection(), self.node(), self.container(), port)
        tunnel.connect_tcpip_direct(self)  # gets the tunnel to send the data here instead of creating tcp sockets
        self.process = tunnel
        self.location().tunnels[tunnel.uuid] = tunnel  # needs this so 'close proxy' events are handled properly
        self.loop().register_exclusive(self.paramiko_channel.fileno(), self.event,
                                       comment="SSH Tunnel " + self.process.uuid.decode())

    # flows

    def data(self, obj, data):
        # from remote
        # if len(data) > 3:
        #     logging.debug("[chan %d] <== %s" % (self.paramiko_channel.get_id(), data.decode()))
        try:
            self.paramiko_channel.sendall(data)
        except OSError:
            logging.debug("[chan %d] Failed to send: %s" % (self.paramiko_channel.get_id(), data.decode()))

    def stderr(self, obj, data):
        logging.debug("[chan %d - stderr] <== %s" % (self.paramiko_channel.get_id(), data.decode()))
        try:
            self.paramiko_channel.sendall_stderr(data)
        except OSError:
            logging.debug("[chan %d - stderr] Failed to send: %s" % (self.paramiko_channel.get_id(), data.decode()))

    def event(self, fd):
        # A file descriptor is triggered on 'our' side
        while self.paramiko_channel.recv_ready():
            data = self.paramiko_channel.recv(8192)  # fits inside a jumbo frame at the other end
            # if len(data) > 3:
            #     logging.debug("[chan %d] ==> %s" % (self.paramiko_channel.get_id(), data.decode()))
            self.process.stdin(data)


class SshReverse:
    socat_re = re.compile(b'^\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2} socat')

    def __init__(self, dest_addr, container, transport):
        # init
        self.dest_addr = dest_addr
        self.container = weakref.ref(container)
        self.transport = weakref.ref(transport)  # needed for open_forwarded_tcpip_channel
        self.process_channel = {}
        self.listening_process = self.spawn_socat()

    def spawn_socat(self):
        # spawn the process that will receive connections inside the container
        logging.debug("Spawning remote listener process for port: " + str(self.dest_addr[1]))
        proc = 'socat -ls -d -d TCP4-LISTEN:%d,reuseaddr STDIO' % self.dest_addr[1]
        return self.container().spawn_process(proc,
                                              data_callback=self.data,
                                              termination_callback=self.terminated,
                                              stderr_callback=self.stderr)

    def spawn_channel(self):
        # does the channel exist already? (set up as part of pending)
        if self.listening_process in self.process_channel:
            return

        # is this a broken reverse?
        if self.listening_process is None:
            self.fail_words("Cannot spawn channel because port is already being listened: " + str(self.dest_addr[1]))
            return

        # when paramiko has finished doing it's thing, we can spawn a forwarded channel
        src_addr = ('127.0.0.1', self.dest_addr[1])
        try:
            p_channel = self.transport().paramiko_transport.open_forwarded_tcpip_channel(src_addr, self.dest_addr)
        except paramiko.ssh_exception.ChannelException:
            self.fail_words("Failed to connect reverse channel onto port: " + str(self.dest_addr[1]))
            return
        channel = SshChannel(p_channel, self.container())

        # hook into the message loop
        try:
            channel.loop().register_exclusive(p_channel.fileno(),
                                              channel.event,  # sent directly from channel, does not filter through here
                                              comment="SSH Reverse " + self.listening_process.uuid.decode())
        except RuntimeError:
            self.fail_words("Tried to register exclusive twice, closing channel: " + str(p_channel.get_id()))
            channel.close(self, 1)
            return

        # hook up indexes
        self.process_channel[self.listening_process] = channel
        channel.process = self.listening_process
        logging.debug("[chan %d] opened to port: %d" % (channel.get_id(), self.dest_addr[1]))
        return channel

    def terminated(self, process, returncode):
        # single terminated process
        logging.debug("SshReverse process terminated: %s (%d)" % (process.uuid.decode(), returncode))

        # may be an established channel, not just the listening process
        try:
            channel = self.process_channel[process]
            channel.close(process, returncode)
            del self.process_channel[process]
        except KeyError:
            logging.debug("...did not have an associated channel")

    def close(self):
        logging.debug("Closing processes for SSH reverse tunnel on port: " + str(self.dest_addr[1]))
        for process in list(self.process_channel.keys()):
            process.destroy()
        if self.listening_process is not None:
            self.listening_process.destroy()

    # Data from the actual process
    def data(self, process, data):
        self.process_channel[process].data(process, data)

    # When socat accepts the connection, we need to create it on this side
    def stderr(self, process, data):

        # see if this is normal stderr
        if SshReverse.socat_re.match(data) is None:
            self.process_channel[process].stderr(process, data)
            return

        # we caught something from socat
        logging.debug("(socat) " + data.decode()[:-1])
        if b'accepting connection from' in data:
            self.spawn_channel()
            self.listening_process = self.spawn_socat()  # spawn another listener
            return

        if b'Address already in use' in data:
            # going to let process termination actually cause the whole thing to shut down
            self.fail_words('Cannot open forwarding channel, port is in use: ' + str(self.dest_addr[1]))
            return

    def fail_words(self, words):
        # echo the failure message down the ssh session's lead channel if possible
        self.transport().lead_channel.stderr(self, words.encode() + b'\r\n')
        logging.warning(words)



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import shortuuid
import weakref
from typing import Optional, List, Union, Callable, Any, Tuple
from . import Waitable, Killable, Connectable
from .tunnel import Tunnel
from .process import Process
from .ssh import SshServer


class Container(Waitable, Killable, Connectable):
    """An object representing a single container. Do not instantiate directly, use node.spawn."""
    def __init__(self, parent, image, uuid, docker_config, env, volumes, *, stdout_callback, termination_callback):
        Waitable.__init__(self)
        Killable.__init__(self)
        Connectable.__init__(self, parent.conn(), uuid, parent, None)
        self.parent = weakref.ref(parent)
        self.location = weakref.ref(self.parent().parent())
        self.image = image
        self.processes = {}
        self.ssh_servers = {}
        self.docker_config = docker_config
        self.env = env
        self.volumes = volumes
        self.stdout_callback = stdout_callback
        self.termination_callback = termination_callback

    def ip(self):
        """Reports the container's ip address"""
        self.ensure_alive()
        self.wait_until_ready()
        return self.ip

    def stdin(self, data: bytes):
        """Writes the data into the container's stdin.

        :param data: The data to be written."""
        self.ensure_alive()
        self.wait_until_ready()
        self.conn().send_cmd(b'stdin_container', {'node': self.parent().pk,
                                                  'container': self.uuid},
                             bulk=data)

    def attach_tunnel(self, dest_port: int, localport: Optional[int]=None, bind: Optional[str]=None) -> Tunnel:
        """Creates a TCP proxy between localhost and a container.

        :param dest_port: The TCP port on the container to connect to.
        :param localport: Optional choice of local port no.
        :param bind: Optionally bind to an address other than localhost.
        :return: A Tunnel object.

        This call does no checking to ensure the server side is ready -
        but a failed connection will not destroy the tunnel itself and hence it can be used for polling.
        If the optional local port number is left as default, one will be automatically chosen.
        """
        self.ensure_alive()
        localport = dest_port if localport is None else localport
        return self.location()._tunnel_onto(self, dest_port, localport, bind)

    def wait_http_200(self, dest_port: Optional[int]=80, fqdn: Optional[str]='localhost', path: Optional[str]='') \
            -> Tunnel:
        """Poll until an http 200 is returned.

        :param dest_port: Override the default port.
        :param fqdn: A host name to use in the http request.
        :param path: A path on the server - appended to /
        :return: A Tunnel object.
        """
        self.ensure_alive()
        self.wait_until_ready()
        return self.location()._wait_http_200(self, dest_port, fqdn, path)

    def destroy_tunnel(self, tunnel: Tunnel):
        """Destroy a tunnel

        :param tunnel: The tunnel to be destroyed."""
        if not isinstance(tunnel, Tunnel):
            raise TypeError()
        self.ensure_alive()
        self.location()._destroy_tunnel(tunnel, container=self)

    def all_tunnels(self) -> List[Tunnel]:
        """Returns all the tunnels connected to this container

        :return: A list of Tunnel objects"""
        self.ensure_alive()
        return [t for t in self.location().tunnels.values() if t.container() == self]

    def allow_connection_from(self, container: 'Container'):
        """Allow another container to call this one over private ip

        :param container: The container that will be allowed to call."""
        self.ensure_alive()
        self.wait_until_ready()
        super().allow_connection_from(container)

    def disallow_connection_from(self, container: 'Container'):
        """Stop allowing another container to call this one over private ip

        :param container: The container that will no longer be allowed to call."""
        if self.bail_if_dead():
            return
        super().disallow_connection_from(container)

    def spawn_process(self, remote_command: str,
                      data_callback: Optional[Callable]=None,
                      termination_callback: Optional[Callable]=None,
                      stderr_callback: Optional[Callable]=None) -> Process:
        """Spawn a process within a container, receives data asynchronously via a callback.

        :param remote_command: The command to remotely launch as a string (i.e. not list).
        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param termination_callback: For when the process completes - signature (object), returncode.
        :param stderr_callback: For data emitted by the process's stderr stream - signature (object, bytes).
        :return: A Process object.
        """
        if isinstance(remote_command, list):
            raise ValueError("Pass the command as a single string (that gets passed to a shell), not a list")
        self.ensure_alive()
        self.wait_until_ready()

        # Cool, go.
        logging.info("Container (%s) spawning process: '%s'" % (self.uuid.decode(), remote_command))

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        spawn_command_uuid = shortuuid.uuid().encode()
        self.processes[spawn_command_uuid] = Process(self, spawn_command_uuid,
                                                     data_callback,
                                                     termination_callback,
                                                     stderr_callback)
        self.conn().send_cmd(b'spawn_process', {'node': self.parent().pk,
                                                'container': self.uuid,
                                                'command': remote_command},
                             uuid=spawn_command_uuid,
                             reply_callback=self._process_callback)
        logging.info("...process id: " + spawn_command_uuid.decode())
        return self.processes[spawn_command_uuid]

    def run_process(self, remote_command: str, *,
                    data_callback: Optional[Callable]=None,
                    stderr_callback: Optional[Callable]=None) -> (bytes, bytes, str):
        """Run a process once, synchronously, without pty.

        :param remote_command: The command to run remotely.
        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param stderr_callback: For data emitted by the process's stderr stream - signature (object, bytes).
        :return: stdout from the process, stderr from the process, exit code (as string).
        """
        if isinstance(remote_command, list):
            raise ValueError("Pass single shot commands a string.")
        self.ensure_alive()
        self.wait_until_ready()

        # Cool, go.
        logging.info("Container (%s) running process: '%s'" % (self.uuid.decode(), remote_command))

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        msg = self.conn().send_blocking_cmd(b'run_process', {'node': self.parent().pk,
                                                             'container': self.uuid,
                                                             'command': remote_command})
        if msg is None:
            raise ValueError("Blocking process failed: " + remote_command)
        if data_callback is not None and len(msg.params['stdout']) > 0:
            data_callback(self, msg.params['stdout'])
        if stderr_callback is not None and len(msg.params['stderr']) > 0:
            stderr_callback(self, msg.params['stderr'])
        return msg.params['stdout'], msg.params['stderr'], msg.params['exit_code']

    def spawn_shell(self,
                    data_callback: Optional[Callable]=None,
                    termination_callback: Optional[Callable]=None,
                    echo: bool=False) -> Process:
        """Spawn a shell within a container, expose as a process

        :param data_callback: A callback for arriving data - signature (object, bytes).
        :param termination_callback: For when the process completes - signature (object, returncode).
        :param echo: Whether or not the shell echoes input.
        :return: A Process object."""
        self.ensure_alive()
        self.wait_until_ready()

        # get the node to launch the process for us
        # we need the uuid of the spawn command because it's used to indicate when the process has terminated
        spawn_command_uuid = shortuuid.uuid().encode()
        self.processes[spawn_command_uuid] = Process(self, spawn_command_uuid, data_callback, termination_callback)
        self.conn().send_cmd(b'spawn_shell', {'node': self.parent().pk,
                                              'container': self.uuid,
                                              'echo': echo},
                             uuid=spawn_command_uuid,
                             reply_callback=self._process_callback)
        logging.info("Spawned shell: " + spawn_command_uuid.decode())
        return self.processes[spawn_command_uuid]

    def destroy_process(self, process: Process):
        """Destroy a process

        :param process: The process to be destroyed."""
        if not isinstance(process, Process):
            raise TypeError()
        self.ensure_alive()
        if process.uuid not in self.processes:
            raise ValueError("Process is not apparently a child of this container")
        process.destroy()
        del self.processes[process.uuid]

    def all_processes(self) -> List[Process]:
        """Returns all the processes (that were manually launched) running on this container.

        :return: A list of Process objects"""
        self.ensure_alive()
        return list(self.processes.values())

    def create_ssh_server(self, port: int=2222) -> SshServer:
        """Create an ssh/sftp server on the given port.

        :param port: Local tcp port number
        :return: An SshServer object."""
        self.ensure_alive()
        self.wait_until_ready()
        s = SshServer(self, port)
        s.start()
        self.ssh_servers[s.uuid] = s
        return s

    def destroy_ssh_server(self, server):
        """Destroy an ssh/sftp server than belongs to this container.

        :param server: An SshServer object."""
        self.ensure_alive()
        if not isinstance(server, SshServer):
            raise TypeError()
        if server.uuid not in self.ssh_servers:
            raise ValueError("Server does not belong to this container")
        server.stop()
        del self.ssh_servers[server.uuid]

    def fetch(self, filename: str) -> bytes:
        """Fetch a single file from the container.

        :param filename: The full-path name of the file to be retrieved.
        :return: the contents of the file as a bytes object.

        Since the file gets loaded into memory, this is a bad way to move large files (>1GB)."""
        self.ensure_alive()
        self.wait_until_ready()
        return self.conn().send_blocking_cmd(b'fetch_file', {'node': self.parent().pk,
                                                             'container': self.uuid,
                                                             'filename': filename}).bulk

    def put(self, filename: str, data: bytes):
        """Put a file into the container.

        :param filename: The full-path name of the file to be placed.
        :param data: The contents of the file as a bytes object.

        This will just overwrite so be careful. Note that new file paths are created on demand.
        Similarly to fetch, this is a bad way to move large files (>1GB).
        """
        self.ensure_alive()
        self.wait_until_ready()
        self.conn().send_blocking_cmd(b'put_file', {'node': self.parent().pk,
                                                    'container': self.uuid,
                                                    'filename': filename}, bulk=data)

    def reboot(self, reset_filesystem: Optional[bool]=False):
        """Synchronously reboot a container, optionally resetting the filesystem.

        :param reset_filesystem: Reset the container's filesystem to its 'as booted' state."""
        if self.dead:
            raise RuntimeError("Tried to restart a container but it has already been removed from the node")
        self.ensure_alive()
        self.wait_until_ready()  # in this case means it has been configured, prepared etc. once already

        # Restart
        logging.info("Restarting container: " + self.uuid.decode())
        self.mark_not_ready()  # so calls are forced to wait until the container reports it has rebooted
        self.conn().send_cmd(b'reboot_container', {'node': self.parent().pk,
                                                   'container': self.uuid,
                                                   'reset_filesystem': reset_filesystem},
                             reply_callback=self.parent()._container_status_update)
        self.wait_until_ready()

    def _internal_destroy(self):
        # Destroy this container
        if self.bail_if_dead():
            return
        self.ensure_alive()
        self.wait_until_ready()

        # Fake the processes being destroyed (they will be anyway)
        for proc in list(self.processes.values()):
            proc.destroy(with_command=False)

        # Destroy any tunnels
        for tun in self.all_tunnels():
            self.location()._destroy_tunnel(tun, self, with_command=False)

        # Destroy (async)
        self.conn().send_cmd(b'destroy_container', {'node': self.parent().pk,
                                                    'container': self.uuid})
        self.mark_as_dead()
        logging.info("Destroying container: " + self.uuid.decode())

    def _process_callback(self, msg):
        if self.bail_if_dead():
            return

        if msg.uuid not in self.processes:
            logging.debug("Message arrived for an unknown process: " + msg.uuid.decode())
            return

        # logging.debug("Received data from process: " + msg.uuid.decode())
        # logging.debug(msg.bulk.decode())
        self.processes[msg.uuid].give_me_messages(msg)

    def __repr__(self):
        return "<tfnz.container.Container object at %x (image=%s uuid=%s)>" % (id(self), self.image, self.uuid.decode())


class ExternalContainer(Connectable):
    """An object representing a container managed by another session (and the same user) but advertised using a tag"""

    def __init__(self, conn: 'Connection', uuid: str, node: Union['Node', bytes], ip):
        super().__init__(conn, uuid, node, ip)



# Copyright (c) 2018 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from tfnz.location import Location
from tfnz.container import Container
from tfnz.volume import Volume


class Postgresql:
    @staticmethod
    def spawn(location: Location, volume: Volume, *, log_callback=None) -> Container:
        node = location.node()
        ctr = node.spawn_container('tfnz/postgresql',
                                   sleep=True,
                                   volumes=[(volume, '/var/lib/postgresql')]).wait_until_ready()
        ctr.run_process('openrc', data_callback=log_callback)
        ctr.run_process('touch /run/openrc/softlevel')
        ctr.run_process('openrc-run /etc/init.d/postgresql start',
                        data_callback=log_callback, stderr_callback=log_callback)
        return ctr



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
import re
from datetime import date
from tfnz.location import Location
from tfnz.volume import Volume

dkim_template = '''
    dkim_domain = %s
    dkim_selector = %s
    dkim_private_key = /var/mail/dkim.key.pem
'''


class Mail:
    def __init__(self, domain_name: str, recipients: [str], *, ssh: bool=False, image: str=None, dkim: bool=False):
        """Instantiate a mail server object.
        If dkim=True, will log the necessary records to be added to DNS
        The certificate and key are given as strings (not filenames).
        The one you use for the webserver is fine if it doesn't say 'www' on it.

        :param domain_name: The domain name to serve.
        :param recipients: A list of recipients in "username:password" strings.
        :param ssh: Create an ssh server for debugging on port 2222.
        :param image: Use a container image other than tfnz/mail.
        :param dkim: Sign outgoing emails.
        :return: A dict representation of image metadata."""
        # check the recipients are in name:pass format
        recipient_re = re.compile('\A[a-zA-Z0-9_.+-]+:\S+\Z')
        self.recipients = [r.strip('\n') for r in recipients]
        if len(self.recipients) == 0:
            raise ValueError("Mail needs to be passed at least one recipient")
        for recipient in self.recipients:
            if recipient_re.match(recipient) is None:
                raise ValueError("Email addresses should be in a 'username:password' format")
        if domain_name is None or len(domain_name) == 0:
            raise ValueError("Exim needs a domain name to function (try 'local'?)")
        self.domain_name = domain_name
        self.ssh = ssh
        self.server_condition = '${if !eq{$tls_cipher}{}}'  # nasty hack to get around .format trying to expand it
        self.system_aliases_data= '${lookup{$local_part}lsearch{/etc/aliases}}'
        self.dkim = dkim
        self.dkim_text = ''
        self.image = 'tfnz/mail' if image is None else image

    def spawn(self, location: Location, volume: Volume, *, log_callback=None,
              local_smtp=25, local_smtps=465, local_imap=993):
        """Instantiate the mail server container.
        Anything with a username/password has to go over SSL. So port 25 is only for mail delivery, not relay.

        :param location: A location (object) to connect to.
        :param volume: A volume (object) to use as a persistent store.
        :param log_callback: An optional callback for log messages -  - signature (object, bytes)
        :param local_smtp: TCP port to open locally for smtp traffic - can only be used for delivery.
        :param local_smtps: TCP port to open locally for smtp (with SSL) traffic.
        :param local_imap: TCP port to open locally for IMAP (with SSL).
        """
        # create the container
        node = location.node()
        container = node.spawn_container(self.image,
                                         volumes=[(volume, '/var/mail')],
                                         termination_callback=location.disconnect,
                                         sleep=True).wait_until_ready()

        # maybe an ssh window
        if self.ssh:
            container.create_ssh_server()

        # create certs if not there
        if b'server.pem' not in container.run_process('ls /var/mail')[0]:
            container.run_process('openssl req -x509 -newkey rsa:1024 -sha256 '
                                  '-keyout /var/mail/server.key -out /var/mail/server.pem -nodes -days 9999 '
                                  '-subj \'/CN=%s\'' % self.domain_name)

        if self.dkim:
            # same again for dkim
            if b'dkim.key.pem' not in container.run_process('ls /var/mail')[0]:
                container.run_process('openssl genrsa -out /var/mail/dkim.key.pem 1024 -outform PEM')
                container.run_process('openssl rsa -in /var/mail/dkim.key.pem'
                                      '            -out /var/mail/dkim.pub.pem -pubout -outform PEM')
                container.put('/var/mail/dkim_selector', date.today().strftime('%Y%M%d').encode())

            # render the dkim configuration
            dkim_selector = container.fetch('/var/mail/dkim_selector')
            dkim_pk = container.fetch('/var/mail/dkim.pub.pem')
            self.dkim_text = dkim_template % (self.domain_name, dkim_selector.decode())

            # log the DNS changes
            print("@ TXT \"v=spf1 a mx ip4:this.is.mx.ip\"")
            print("%s._domainkey TXT \"v=DKIM1; p=%s\"" %
                  (dkim_selector.decode(), dkim_pk.decode()[27:-26].replace('\n', '')))

        # set hostname
        container.run_process('hostname mail')

        # create user accounts
        passwd_file = ''
        first_recipient = None
        for recipient in self.recipients:
            username, password = recipient.split(':')
            if first_recipient is None:
                first_recipient = username
            pw = container.run_process('doveadm pw -u %s -p %s' % (username, password))[0][:-1]
            passwd_file += '%s:%s:90:101\n' % (username, pw.decode())
        container.put('/etc/dovecot/passwd', passwd_file.encode())

        # create system aliases - uses the first recipient as the destination
        aliases = ['root', 'postmaster', 'hostmaster', 'mail', 'admin', 'abuse',
                   'webmaster', 'mailer-daemon', 'nobody', 'www', 'security']
        alias_file = ''
        for alias in aliases:
            alias_file += '%s: %s\n' % (alias, first_recipient)
        container.put('/etc/aliases', alias_file.encode())

        # render the exim configuration
        exim_template = container.fetch('/home/admin/exim_template').decode()
        container.put('/etc/exim/exim.conf', exim_template.format(self).encode())

        # attach an SMTP tunnel and run exim
        container.spawn_process('tail -f /var/log/exim/mainlog', data_callback=log_callback)
        container.attach_tunnel(25, localport=local_smtp)
        container.attach_tunnel(465, localport=local_smtps)
        container.spawn_process('/usr/sbin/exim -bdf -q30m', data_callback=log_callback)

        # same again with dovecot
        container.attach_tunnel(993, localport=local_imap)
        container.spawn_process('/usr/sbin/dovecot', data_callback=log_callback)



"""
Copyright (c) 2017 David Preece, All rights reserved.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""
import sys
import argparse
import termios  # this *is* used and *can* be found
import tty
import select
import re
import logging
from subprocess import check_call, CalledProcessError
from messidge import default_location
from tfnz.location import Location


def base_argparse(progname, location=True):
    parser = argparse.ArgumentParser(prog=progname)
    if location:
        connection_group = parser.add_argument_group('connection options')
        connection_group.add_argument('--location', help='use a non-default location', metavar='x.20ft.nz')
        connection_group.add_argument('--local', help='a non-dns ip for the location', metavar='x.local')
    return parser


def generic_cli(parser, implementations, *, quiet=True, location=True):
    args = parser.parse_args()
    dl = None
    loc = None

    if 'command' not in args:
        args.command = None

    if args.command is None and None not in implementations:
        parser.print_help()
        return

    # construct the location
    if location:
        try:
            dl = default_location(prefix="~/.20ft") if args.location is None else args.location
        except RuntimeError:
            print("There does not appear to be a 20ft account on this machine.", file=sys.stderr)
            sys.exit(1)

    # go
    try:
        if location:
            loc = Location(dl, location_ip=args.local, quiet=quiet)
        else:
            if not quiet:
                logging.basicConfig(level=logging.DEBUG)
        implementations[args.command](loc, args)
    except ValueError as e:
        print(str(e))
        exit(1)
    except KeyboardInterrupt:
        exit(1)
    finally:
        if loc is not None:
            loc.disconnect()
    exit(0)


# removes a flagged parameter from argv
# note: mutates argv
def remove_flagged(param, argv):
    flag_find = [a for a in enumerate(argv) if a[1] == param]
    if len(flag_find) == 1:
        del argv[flag_find[0][0]]  # flag
        del argv[flag_find[0][0]]  # and the value


def systemd(location, args, argv, preboot, cert):
    # note that this is *knee* *deep* in potential security holes but since it is connecting to
    # a non-multi-tenanted server under the user's control I'm not all that bothered about it.
    # you were warned

    # checks
    if '/' not in args.source:
        print("Use a tagged image (i.e. my/example) to create a service")
        return 1
    if '@' not in args.systemd:
        print("Please use a 'user@server' ssh connection string.")
    username = args.systemd.split('@')[0]

    # build the ssh command line
    if args.identity is None:
        ssh = ['ssh', args.systemd]
        sftp = 'sftp %s' % args.systemd
    else:
        ssh = ['ssh', '-i', args.identity, args.systemd]
        sftp = 'sftp -i %s %s' % (args.identity, args.systemd)

    # ensure file structure on the receiving machine
    service_name = args.source.replace('/', '-')
    path = '/home/%s/%s/' % (username, service_name)
    check_call(ssh + ['mkdir', '-p', path])

    # ensure the image is uploaded to the *location* (not the client-local server)
    location.ensure_image_uploaded(args.source)

    # copy any preboot files
    for source, _ in preboot:
        check_call('echo "put %s" | %s:%s' % (source, sftp, path), shell=True)

    # copy a certificate
    if cert is not None:
        check_call('echo "put %s" | %s:%s' % (cert[0], sftp, path), shell=True)
        check_call('echo "put %s" | %s:%s' % (cert[1], sftp, path), shell=True)

    # remove the procname, --systemd and --identity from argv
    del argv[0]
    remove_flagged('--systemd', argv)
    remove_flagged('--identity', argv)

    # render the unit file
    filename = service_name + '.service'
    with open(filename, 'w') as f:
        f.write('''
[Unit]
Description=20ft-%s

[Service]
Type=simple
Environment=PYTHONUNBUFFERED=1
ExecStart=/usr/local/bin/tf %s
WorkingDirectory=%s
KillSignal=SIGINT
TimeoutStopSec=5
Restart=always
RestartSec=5
User=%s
Group=%s

[Install]
WantedBy=multi-user.target
    ''' % (args.source, ' '.join(argv), path, username, username))
    check_call('echo "put %s" | %s:%s\n' % (filename, sftp, path), shell=True)
    check_call(['rm', filename])

    # let systemd know
    try:
        check_call(ssh + ['sudo', 'ln', '-s', path + filename, '/etc/systemd/system/'])
    except CalledProcessError as e:
        print("WARNING: creating symlink into /etc/systemd/system failed - link may well be there already")
    check_call(ssh + ['sudo', 'systemctl', 'enable', service_name])
    check_call(ssh + ['sudo', 'systemctl', 'daemon-reload'])
    check_call(ssh + ['sudo', 'systemctl', 'start', service_name])


class Interactive:
    def __init__(self, loc):
        self.shown_escape_info = False
        self.term_attr = None
        try:
            self.term_attr = termios.tcgetattr(sys.stdout.fileno())
        except termios.error:
            pass
        self.location = loc
        self.running = True

    def stdin_loop(self, container):
        # do we need to let the user know?
        if not self.shown_escape_info:
            print("Interactive session - escape is triple \'^]\'.")
            self.shown_escape_info = True

        tty.setraw(sys.stdin.fileno())
        while self.running:
            ready = select.select((sys.stdin,), (), (), 0.5)
            if len(ready[0]) != 0:
                data = sys.stdin.read(1)
                container.stdin(data.encode())

    def stdout_callback(self, ctr, out):
        # strip nasty control code things
        parts = re.split(b'\x1b\[\d*n', out, maxsplit=1)
        sys.stdout.buffer.write(parts[0] + (parts[1] if len(parts) > 1 else b''))
        sys.stdout.flush()

    def termination_callback(self, ctr, returncode):
        self.running = False
        if self.term_attr is not None:
            termios.tcsetattr(sys.stdout.fileno(), termios.TCSANOW, self.term_attr)



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import hashlib
import logging
import io
import lzma
from tarfile import TarFile, ReadError
from .docker import Docker


class Sender:

    @staticmethod
    def layer_stack(descr):
        """Returns a list of the layers necessary to create the passed docker image id."""
        # find layers
        layers = descr['RootFS']['Layers']

        # take the layers and prevent the same layer being applied twice in a row
        single_run_layers = []
        last_layer = None
        for layer in layers:
            layer = layer[7:]
            if layer == last_layer:
                continue
            single_run_layers.append(layer)
            last_layer = layer
        return single_run_layers

    @staticmethod
    def upload_requirements(layers, conn):
        # ask the location what it needs
        return conn.send_blocking_cmd(b'upload_requirements', {'layers': layers}).params

    @staticmethod
    def send(docker_image_id, layers, conn):
        """Internal use: Send the missing layers to the location."""
        if len(layers) == 0:
            logging.info("No layers need uploading for: " + docker_image_id)
            return

        # get docker to export *all* the layers (not like we have a choice, would be happy to be informed otherwise)
        # note that making a fake registry was tried and found to be horrible
        logging.info("Waiting for Docker to export image...")
        tarball = Docker.tarball(docker_image_id)
        try:
            raw_top_tar = io.BytesIO(tarball)
            top_tar = TarFile(fileobj=raw_top_tar)
        except ReadError:
            raise RuntimeError("Local docker does not appear to have image: " + docker_image_id)

        # sha256 and send until all our requirements are met
        for member in top_tar.getmembers():
            # only even remotely interested in the layers
            logging.debug("Examining: " + str(member))
            if '/layer.tar' not in str(member):
                continue

            # extract and hash the data
            layer_data = top_tar.extractfile(member).read()
            sha256 = hashlib.sha256(layer_data).hexdigest()

            # is this one we care about?
            if sha256 in layers:
                logging.info("Uploading: " + sha256[:16])

                # send in compressed 4MB chunks
                slab_size = 4*1024*1024
                data_loc = 0
                slab = 0
                data_length = len(layer_data)
                logging.info("Uploading slabs: " + str((data_length // slab_size) + 1))
                while data_loc < data_length:
                    end_byte = (slab+1) * slab_size
                    if end_byte >= len(layer_data):
                        end_byte = len(layer_data)
                    send_data = lzma.compress(layer_data[data_loc:end_byte], preset=1)
                    conn.send_cmd(b'upload_slab', {'sha256': sha256, 'slab': slab}, bulk=send_data)
                    data_loc += slab_size
                    slab += 1
                    logging.info("Sending slab: " + str(slab))

                # this is the end
                # the upload_complete call can take ages to happen because it'll be behind all the slabs
                msg = conn.send_blocking_cmd(b'upload_complete', {'sha256': sha256, 'slabs': slab}, timeout=300)
                logging.info(msg.params['log'])




# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import re
import weakref
import shortuuid
from messidge.client.connection import Connection
from typing import Optional, Union, List
from base64 import b64encode
from _thread import allocate_lock


class Waitable:
    """An object that can be waited on (until marked as ready)"""
    def __init__(self, locked: Optional[bool]=True):
        self.wait_lock = allocate_lock()
        self.exception = None
        if locked:
            self.wait_lock.acquire()

    def __del__(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def wait_until_ready(self, timeout: Optional[int]=30):
        """Blocks waiting for a (normally asynchronous) update indicating the object is ready.

        Note this also causes exceptions that would previously have been raised on the background thread
        to be raise on the calling (i.e. main) thread."""
        # this lock is used for waiting on while uploading layers, needs to be long
        self.wait_lock.acquire(timeout=timeout)
        self.wait_lock.release()
        if self.exception:
            raise self.exception
        return self

    def mark_as_ready(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def mark_not_ready(self):
        if not self.wait_lock.locked():
            self.wait_lock.acquire()

    def unblock_and_raise(self, exception):
        # releases the lock but causes the thread in 'wait_until_ready' to raise an exception
        self.exception = exception
        self.mark_as_ready()


class Killable:
    """An object that can be marked as dead - and either bail and carry on, or raise if dead"""
    def __init__(self):
        self.dead = False

    def mark_as_dead(self):
        self.dead = True

    def bail_if_dead(self):
        # use this where the user is not really to blame
        if self.dead:
            logging.debug("Object was previously terminated (carrying on): " + self.__repr__())
        return self.dead

    def ensure_alive(self):
        # use this where the user has tried to use the object but it's kinda their fault it's no longer there
        if self.dead:
            raise ValueError("Cannot call method on a terminated object: " + self.__repr__())


class Connectable:
    """A resource that can be connected to in the private IP space"""
    def __init__(self, conn: Connection, uuid: str, node, ip):
        self.conn = weakref.ref(conn)
        self.uuid = uuid
        self.node_pk = node if isinstance(node, bytes) else node.pk
        self.ip = ip  # gets used externally, don't delete it

    def allow_connection_from(self, obj):
        self.conn().send_blocking_cmd(b'allow_connection', {'node': self.node_pk,
                                                            'container': self.uuid,
                                                            'ip': obj.ip})
        logging.info("Allowed connection (from %s) on: %s" % (obj.uuid.decode(), self.uuid.decode()))

    def disallow_connection_from(self, obj):
        self.conn().send_cmd(b'disallow_connection', {'node': self.node_pk,
                                                      'container': self.uuid,
                                                      'ip': obj.ip})
        logging.info("Disallowed connection (from %s) on: %s" % (obj.uuid.decode(), self.uuid.decode()))


class Taggable:
    """A resource that might have a tag - namespaced by user pk."""
    # user, uuid and tag are all *held* as bytes
    tag_re = re.compile(b'\A[^0-9a-z\-_.]*\Z')
    short_uuid_re = re.compile(b'\A[' + shortuuid.get_alphabet().encode() + b']{22}\Z')

    def __init__(self, user: Union[str, bytes], uuid: Union[str, bytes], tag: Optional[Union[str, bytes]]=None):
        if user is None or uuid is None:
            raise RuntimeError('Taggable resources must be constructed with at least a pk and uui')
        if isinstance(user, str):
            self.user = user.encode()
        else:
            self.user = user

        if isinstance(uuid, str):
            self.uuid = uuid.encode()
        else:
            self.uuid = uuid

        self.tag = None if tag is None else Taggable.valid_tag(tag)

    def uuid_key(self) -> (bytes, bytes):
        return self.user, self.uuid

    def tag_key(self) -> (bytes, bytes):
        """Effectively a namespaced tag."""
        if self.tag is None:
            return None
        return self.user, self.tag

    @staticmethod
    def valid_tag(tag: Union[bytes, str]) -> Union[bytes, None]:
        """Ensure that the passed tag is at least vaguely plausible - does not check for clashes"""
        if tag is None:
            return None
        if len(tag) == 0:
            raise ValueError("Tag passed for approval was blank")
        tag = tag.lower()
        if isinstance(tag, str):
            tag = tag.encode()
        if Taggable.tag_re.search(tag) is not None:
            raise ValueError("Tag names can only use 0-9 a-z - _ and .")
        if Taggable.short_uuid_re.match(tag) is not None:
            raise ValueError("Tag names cannot look like UUIDs")
        return tag

    def namespaced_display_name(self) -> str:
        return self.uuid.decode() if self.tag is None else (self.uuid.decode() + ':' + self.tag.decode())

    def global_display_name(self) -> str:
        return b64encode(self.user).decode() + ':' + self.namespaced_display_name()


class TaggedCollection:
    """A collection of taggable objects"""
    def __init__(self, initial: Optional[List[Taggable]]=None):
        self.objects = {}
        self.uuid_uuidkey = {}
        self.uniques = 0
        if initial is not None:
            for init in initial:
                self.add(init)

    def __del__(self):
        for obj in self.objects.values():
            del obj
        self.objects = {}

    # emulating a dictionary
    def __len__(self) -> int:
        return self.uniques

    def __getitem__(self, uuid: Union[str, bytes]) -> Taggable:  # applies to just UUID's
        if isinstance(uuid, str):
            uuidkey = self.uuid_uuidkey[uuid.encode()]
        else:
            uuidkey = self.uuid_uuidkey[uuid]
        return self.objects[uuidkey]

    def __contains__(self, uuid: Union[str, bytes]) -> bool:
        try:
            self[uuid]  # throws if it can't get it so, yes, this does actually do something
            return True
        except KeyError:
            return False

    def __setitem__(self, key, value):
        raise RuntimeError("Tagged collection can only be inserted to with the 'add' method")

    def __iter__(self):
        raise RuntimeError("Cannot iterate over a tagged collection")

    # what we came here for
    def add(self, obj: Taggable):
        if self.will_clash(obj.user, obj.uuid, obj.tag):
            raise RuntimeError("Cannot add to TaggedCollection because there will be a namespace clash")
        self.objects[obj.uuid_key()] = obj
        self.uuid_uuidkey[obj.uuid] = obj.uuid_key()
        if obj.tag_key() is not None:
            self.objects[obj.tag_key()] = obj
        self.uniques += 1

    def get(self, user: Union[bytes, str], key: Union[bytes, str]) -> Taggable:
        """Fetch using an ill-defined key: uuid or tag or uuid:tag"""
        if key is None:
            raise RuntimeError("Key not passed when fetching from TaggedCollection")
        if isinstance(user, str):
            user = user.encode()
        if isinstance(key, str):
            key = key.encode()

        # uuid or key or uuid:key?
        parts = key.split(b':')
        if len(parts) > 2:
            raise ValueError("Too many parts in tag: " + key.decode())

        # will match (user, uuid) or (user, tag)
        try:
            return self.objects[(user, parts[0])]
        except KeyError:
            # it seems as if we would need to search using parts[1] but...
            # we have both (user, uuid) and (user, tag) in the collection so it doesn't matter if we've passed
            # uuid, uuid:tag (matches UUID), or tag
            raise KeyError('Failed to get from a TaggedCollection with user=%s key=%s' % (user, key))

    def remove(self, obj: Taggable):
        del self.objects[obj.uuid_key()]
        del self.uuid_uuidkey[obj.uuid]
        if obj.tag_key() in self.objects:
            del self.objects[obj.tag_key()]
        self.uniques -= 1
        
    def will_clash(self, user: bytes, uuid: bytes, tag: bytes) -> bool:
        if tag is not None:
            if (user, tag) in self.objects:
                return True
        if (user, uuid) in self.objects:
            return True
        return False

    def values(self):
        return set(self.objects.values())  # de-dupe
